# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of django-json-dbindex nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
import logging
import sys
from django.db import connections


def execute_raw(sql, database='default'):
    """Execute a raw SQL command

    sql (string) : SQL command
    database (string): the database name configured in settings
    """
    try:
        cursor = connections[database].cursor()
        cursor.execute(sql)
        cursor.close()
        return 0
    except:
        logging.error('Cant execute %s' % (sql))
        sys.stderr.write('Cant execute [%s]\n' % (sql))
        sys.exit(1)


def lookup(sql, params=None, database='default'):
    """Execute a raw SQL command

    sql (string) : SQL command
    database (string): the database name configured in settings
    """
    try:
        cursor = connections[database].cursor()
        cursor.execute(sql, params)
        row = cursor.fetchone()
        cursor.close()
        return row[0]
    except:
        logging.error('Cant execute %s' % (sql))
        sys.stderr.write("%s\n" % (mogrify(sql, params)))
        return False


def mogrify(sql, params=None, database='default'):
    """Mogrify a command

    sql (string) : SQL command
    database (string): the database name configured in settings
    """
    cursor = connections[database].cursor()
    try:
        return cursor.mogrify(sql, params)
    except AttributeError:
        # SQLite has no mogrify function
        return "%s -> %s" % (sql, params)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of django-json-dbindex nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
import dbcommands
import sys


ACTIONS = ['insert', 'update', 'delete']

SQLITE_BACKEND = 'django.db.backends.sqlite3'
PG_BACKEND = 'django.db.backends.postgresql_psycopg2'


class DatabaseNotSupported(Exception):
    pass


class DatabaseUnknown(Exception):
    pass


class ActionUnknown(Exception):
    pass


def function_name(table, column, action):
    return "{0}_{1}_{2}()".format(table, column, action)


def trigger_name(table, column, action):
    return "{0}_{1}_{2}_trigger".format(table, column, action)


def triggers_name(table, column):
    tgs = []
    for action in ACTIONS:
        tgs.append("{0}_{1}_{2}_trigger".format(table, column, action))
    return tgs


def table_name(table, column):
    """Compute the table name

    table (string) : the originate table's name
    column : the column name

    return : string
    """
    return "{0}__{1}_agg".format(table, column)


class AggTrigger(object):
    """Manage database triggers
    """

    def __init__(self, engine, table, column,
                 aggregats=['count'],
                 database='default'):
        self.verbose = 0
        self.table = table
        self.column = column
        self.aggregats = aggregats
        self.database = database

        # use the right backend
        if engine == PG_BACKEND:
            from databases.pg import TriggerPostgreSQL
            self.backend = TriggerPostgreSQL()
        else:
            raise DatabaseNotSupported()

    @property
    def table_name(self):
        return table_name(self.table, self.column)

    @property
    def triggers_name(self):
        return triggers_name(self.table, self.column)

    @property
    def functions_name(self):
        fns = []
        for action in ACTIONS:
            fns.append(function_name(self.table, self.column, action))
        return fns

    def create_objects(self):
        """Create all needed objects
        table (string)
        column (string)
        aggregats (array)
        """
        res = self.create_table()
        res = res + self.create_functions()
        res = res + self.create_triggers()
        return res

    def create_table(self):
        """
        table (string)
        column (string)
        aggregats (array)

        Return :
        - 0 is case of creation + success
        - 1 if table already exists
        """
        if not self.agg_table_ispresent():
            sql = self.sql_create_table()
            return dbcommands.execute_raw(sql, self.database)
        else:
            return 1

    def create_triggers(self):
        """Create all triggers
        """
        res = 0
        for (trigger_name, sql) in self.sql_create_triggers():
            if res == 0:
                if not self.trigger_on_table_is_present(trigger_name):
                    res = self.create_trigger(sql)
        return res

    def create_trigger(self, sql):
        """Create a trigger
        """
        if self.verbose > 2:
            stm = dbcommands.mogrify(sql, database=self.database)
            sys.stdout.write(stm)
        res = dbcommands.execute_raw(sql, self.database)
        return res

    def create_functions(self):
        """Create all functions in the database
        """
        res = 0
        for sql in self.sql_create_functions():
            if res == 0:
                res = dbcommands.execute_raw(sql, self.database)
        return res

    def sql_create_table(self):
        """
        """
        typname = self.column_typname()
        return self.backend.sql_create_table(self.table_name,
                                             (self.column, typname),
                                             self.aggregats)

    def column_typname(self):
        """Lookup for a column type

        Depending on backend.sql_get_column_typname()[1] we need to
        parse the result or not
        """
        (qry, parsefunc) = self.backend.sql_get_column_typname()
        params = (self.table, self.column)
        res = dbcommands.lookup(qry,
                                params=params,
                                database=self.database)
        if parsefunc is None:
            return res
        else:
            return eval("self.backend.%s('%s', '%s')" % (parsefunc,
                                                         res,
                                                         self.column))

    def sql_init(self):
        """
        Remplissage de la table technique avec les données existantes
        """
        sql = self.backend.sql_init_aggtable(self.table,
                                             self.column,
                                             self.table_name,
                                             self.aggregats)
        return sql

    def initialize(self):
        """Initialize the agregate table with values from source table
        May be long on huge tables
        """
        sql = self.sql_init()
        res = dbcommands.execute_raw(sql, database=self.database)
        return res

    def get_nb_tuples(self):
        """
        Remplissage de la table technique avec les données existantes
        """
        sql = self.backend.sql_nb_tuples()
        params = (self.table,)
        if self.verbose > 2:
            msg = "[SQL] %s\n" % (dbcommands.mogrify(sql,
                                                     params=params,
                                                     database=self.database))
            sys.stdout.write(msg)
        res = dbcommands.lookup(sql, params=params, database=self.database)
        return res

    def sql_create_triggers(self):
        """Declaration des triggers

        Return : Array of tuples (triggername, SQL_COMMAND)
        """
        sql = []
        for action in ACTIONS:
            function = function_name(self.table, self.column, action)
            sql.append(self.sql_create_trigger(self.table,
                                               action,
                                               function))
        return sql

    def sql_create_trigger(self, table, action, function):
        """
        table (string): table name
        action (string): insert, update or create
        function (string): function's name called by the trigger
        """
        tgname = trigger_name(table, self.column, action)
        return (tgname, self.backend.sql_create_trigger(tgname,
                                                        function,
                                                        table,
                                                        action))

    def drop_table(self):
        sql = self.sql_drop_table()
        return dbcommands.execute_raw(sql)

    def drop_functions(self):
        """DROP all related functions
        """
        res = 0
        for sql in self.sql_drop_functions():
            res = res + dbcommands.execute_raw(sql)
        return res

    def sql_drop_functions(self):
        """Return SQL Statements to DROP all functions
        """
        sql = []
        for action in ACTIONS:
            fname = function_name(self.table, self.column, action)
            sql.append(self.sql_drop_function(fname))
        return sql

    def sql_create_functions(self):
        """Create all functions

        table (string)
        column (string)
        aggregats (array)
        """
        sql = []
        for action in ACTIONS:
            sql.append(self.sql_create_function(self.table,
                                                self.column,
                                                self.aggregats,
                                                action))
        return sql

    def sql_create_function(self, table, column, aggregats, action):
        """
        table (string)
        column (string)
        aggregats (array)
        action (string)
        """
        fname = function_name(table, column, action)
        tname = table_name(table, column)
        sql = None
        if action == 'insert':
            sql = self.backend.sql_create_function_insert(fname, table,
                                                          column, tname)
        elif action == 'update':
            sql = self.backend.sql_create_function_update(fname, column,
                                                          tname, action)
        elif action == 'delete':
            sql = self.backend.sql_create_function_delete(fname, column,
                                                          tname, action)
        else:
            raise ActionUnknown(Exception)

        return sql

    def drop_triggers(self, name):
        """DROP related triggers
        """
        res = 0
        for sql in self.sql_drop_triggers(name):
            res = res + dbcommands.execute_raw(sql, self.database)
        return res

    def sql_drop_triggers(self):
        """Return SQL Statements to DROP ALL triggers
        """
        sql = []
        for action in ACTIONS:
            tgname = trigger_name(self.table, self.column, action)
            sql.append(self.sql_drop_trigger(tgname, self.table))
        return sql

    def sql_drop_function(self, fname):
        """Return SQL statement from backend to DROP a function
        """
        return self.backend.sql_drop_function(fname)

    def sql_drop_trigger(self, name, table):
        """Return SQL statement from backend to DROP a trigger
        """
        return self.backend.sql_drop_trigger(name, table)

    def sql_drop_table(self):
        """Return SQL statement from backend to DROP a table
        """
        return self.backend.sql_drop_table(self.table_name)

    def agg_function(self, fname, agg):
        """Return the aggregate name

        fname (string) : min or max
        """
        return self.backend.agg_fname(agg, fname)

    def agg_table_ispresent(self):
        """Check if the agg table is present
        """
        qry = self.backend.sql_table_exists()
        params = (self.table_name,)
        res = dbcommands.lookup(qry,
                                params=params,
                                database=self.database)
        return res

    def triggers_on_table_are_present(self):
        """Check if the triggers are present in database
        """
        res = []
        qry = self.backend.sql_trigger_on_table_exists()
        for trig in self.triggers_name:
            params = (trig, self.table)
            res.append((trig, dbcommands.lookup(qry,
                                                params=params,
                                                database=self.database)))
        return res

    def trigger_on_table_is_present(self, trgname):
        """Check if a trigger is present on a table

        Return False is the trigger does not exists

        trganme : (string) trigger name
        """
        qry = self.backend.sql_trigger_on_table_exists()
        params = (trgname, self.table)
        res = dbcommands.lookup(qry, params=params, database=self.database)
        return res

    def functions_are_present(self):
        """Check if the functions are present in database
        """
        res = []
        qry = self.backend.sql_trigger_function_on_table_exists()
        for func in self.functions_name:
            params = (func[:-2], self.table)
            res.append((func, dbcommands.lookup(qry,
                                                params=params,
                                                database=self.database)))
        return res



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of django-json-dbindex nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
from django.db import models
from django.db import connection
import tool
import util


class TooManyKwargs(Exception):
    pass


class MissingKwargs(Exception):
    pass


class AggTriggManager(models.Manager):

    def optimized_count(self, **kwargs):
        """Return count in optimized manner

        Apple.objects.filter(indice=4).count()
        Apple.objects.optimized_count(indice=4)

        """
        if (len(kwargs) == 0):
            raise MissingKwargs("you must specify the filter clause")
        if (len(kwargs) > 1):
            raise TooManyKwargs()
        # do not care about more than one kwarg
        (key, value) = kwargs.items()[0]
        print "%s = %s" % (key, value)
        (nbe, column, qfilter) = tool.parse_kwarg(key)

        agg = 'agg_count'
        agf = agg.split('_')

        agt = util.AggTrigger(self.model._meta.db_table, column)

        if nbe == 1:
            agf = agg
        elif nbe == 2:
            agf = agt.agg_function(agg, agf[1])

        cursor = connection.cursor()
        qry = """SELECT {} FROM {} WHERE {} %s"""
        tbname = util.table_name(self.model._meta.db_table, column)
        cursor.execute(qry.format(agf, tbname, qfilter), [value])
        print cursor.query
        row = cursor.fetchone()

        return row[0]


class IntegerTriggerField(models.IntegerField):

    description = "An IntegerField with trigger"

    def __init__(self, *args, **kwargs):
        self.aggregate_trigger = ['count']
        super(IntegerTriggerField, self).__init__(*args, **kwargs)


class FloatTriggerField(models.FloatField):

    description = "An FloatField with trigger"

    def __init__(self, *args, **kwargs):
        self.aggregate_trigger = ['count']
        super(FloatTriggerField, self).__init__(*args, **kwargs)



#! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2012 Rodolphe Quiedeville <rodolphe@quiedeville.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import tool


class toolTests(unittest.TestCase):

    def test_equal(self):
        res = tool.parse_kwarg("foo")
        self.assertEqual(res, (1, 'foo', 'foo='))

    def test_greater_than(self):
        res = tool.parse_kwarg("foo__gt")
        self.assertEqual(res, (2, 'foo', 'foo >'))

    def test_greater_or_equal_than(self):
        res = tool.parse_kwarg("foo__gte")
        self.assertEqual(res, (2, 'foo', 'foo >='))

    def test_less_than(self):
        res = tool.parse_kwarg("foo__lt")
        self.assertEqual(res, (2, 'foo', 'foo <'))

    def test_less_or_equal_than(self):
        res = tool.parse_kwarg("foo__lte")
        self.assertEqual(res, (2, 'foo', 'foo <='))



# -*- coding: utf-8 -*-

# Copyright (C) 2014 Rodolphe Quiedeville <rodolphe@quiedeville.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import util
import dbcommands
from mock import MagicMock

PGBACKEND = 'django.db.backends.postgresql_psycopg2'


class utilTests(unittest.TestCase):

    def setUp(self):
        self.agg = util.AggTrigger(PGBACKEND, 'foota', 'foocol', ['agg'])

    def test_init(self):
        self.assertEqual(self.agg.database, 'default')

    def test_agg_table_name(self):
        agg = util.AggTrigger(PGBACKEND, 'foota', 'foocol', ['agg'])
        self.assertTrue(agg.table_name, 'foota_foocol_agg')

    def test_function_name(self):
        self.assertEqual("foo_colfoo_insert()",
                         util.function_name('foo', 'colfoo', 'insert'))

    def test_trigger_name(self):
        self.assertEqual("foo_id_insert_trigger",
                         util.trigger_name('foo', 'id', 'insert'))

    def test_sql_create_trigger(self):
        res = self.agg.sql_create_trigger('foo', 'insert', 'function_called()')
        sql = res[1]
        self.assertEqual(sql[:6], "CREATE")

    def test_sql_create_triggers(self):
        agg = util.AggTrigger(PGBACKEND, 'foota', 'foocol', ['agg'])
        res = agg.sql_create_triggers()
        self.assertEqual(len(res), 3)
        self.assertTrue(isinstance(res, list))

        tgnames = [tgn[0] for tgn in res]
        sqls = [tgn[1] for tgn in res]

        self.assertEqual(tgnames[0], "foota_foocol_insert_trigger")
        self.assertEqual(sqls[0][:6], "CREATE")

    def test_sql_create_function(self):
        agg = util.AggTrigger(PGBACKEND, 'foota', 'foocol', ['agg'])
        res = agg.sql_create_function('book', 'nbpage', ['count'], 'insert')
        self.assertTrue(isinstance(res, unicode))
        self.assertTrue(len(res) > 10)

    def test_sql_create_functions(self):
        agg = util.AggTrigger(PGBACKEND, 'foota', 'foocol', ['agg'])
        res = agg.sql_create_functions()
        self.assertTrue(isinstance(res, list))
        self.assertEqual(len(res), 3)

    def test_sql_drop_functions(self):
        agg = util.AggTrigger(PGBACKEND, 'foota', 'foocol', ['agg'])
        res = agg.sql_drop_functions()
        self.assertTrue(isinstance(res, list))
        self.assertEqual(len(res), 3)

    def test_sql_drop_triggers(self):
        agg = util.AggTrigger(PGBACKEND, 'foota', 'foocol', ['agg'])
        res = agg.sql_drop_triggers()
        self.assertTrue(isinstance(res, list))
        self.assertEqual(len(res), 3)

    def test_sql_create_table(self):
        """Test with only one aggregat
        """
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        res = agg.sql_create_table()
        self.assertTrue(isinstance(res, unicode))
        self.assertTrue(len(res) > 3)

    def test_sql_create_table_twoagg(self):
        """Test with two aggregats
        """
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count', 'min'])
        res = agg.sql_create_table()
        self.assertTrue(isinstance(res, unicode))
        self.assertTrue(len(res) > 3)

    def test_sql_drop_trigger(self):
        res = self.agg.sql_drop_trigger('foo', 'table_bar')
        self.assertTrue(isinstance(res, str))

    def test_sql_drop_function(self):
        res = self.agg.sql_drop_function('foo')
        self.assertTrue(isinstance(res, str))

    def test_sql_drop_table(self):
        res = self.agg.sql_drop_table()
        self.assertTrue(isinstance(res, str))

    def test_agg_create_table(self):
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        dbcommands.execute_raw = MagicMock(return_value=0)
        res = agg.create_table()
        self.assertEqual(res, 0)

    def test_agg_create_triggers(self):
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        dbcommands.execute_raw = MagicMock(return_value=0)
        res = agg.create_triggers()
        self.assertEqual(res, 0)

    def test_agg_create_functions(self):
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        dbcommands.execute_raw = MagicMock(return_value=0)
        res = agg.create_functions()
        self.assertEqual(res, 0)

    def test_agg_create_objects(self):
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        dbcommands.execute_raw = MagicMock(return_value=0)
        res = agg.create_objects()
        self.assertEqual(res, 0)

    def test_agg_sql_init(self):
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        res = agg.sql_init()
        self.assertTrue(isinstance(res, unicode) or isinstance(res, str))

    def test_agg_functions_name(self):
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        res = agg.functions_name
        self.assertTrue(res, ['book_nbpage_insert()',
                              'book_nbpage_update()',
                              'book_nbpage_delete()'])

    def test_functions_are_present(self):
        agg = util.AggTrigger(PGBACKEND, 'book', 'nbpage', ['count'])
        dbcommands.execute_raw = MagicMock(return_value=0)
        res = agg.functions_are_present()

        funcs = [fun[0] for fun in res]

        self.assertEqual(len(res), 3)
        self.assertEqual(funcs, ['book_nbpage_insert()',
                                 'book_nbpage_update()',
                                 'book_nbpage_delete()'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Python packaging."""
import os
import sys

from setuptools import setup


#: Absolute path to directory containing setup.py file.
here = os.path.abspath(os.path.dirname(__file__))
#: Boolean, ``True`` if environment is running Python version 2.
IS_PYTHON2 = sys.version_info[0] == 2


NAME = 'django-aggtrigg'
DESCRIPTION = 'Django complementary index definition and management.'
README = open(os.path.join(here, 'README.rst')).read()
VERSION = open(os.path.join(here, 'VERSION')).read().strip()
AUTHOR = u'Rodolphe Quiédeville'
EMAIL = 'rodolphe@quiedeville.org'
LICENSE = 'BSD'
URL = 'https://django-aggtrigg.readthedocs.org/'
CLASSIFIERS = [
    'Programming Language :: Python :: 2.7',
    'Development Status :: 3 - Alpha',
    'Framework :: Django',
    'License :: OSI Approved :: BSD License',
]
KEYWORDS = [
    'aggregat',
    'trigger',
    'count',
    'min',
    'max'
    'database',
    'index',
    'postgresql',
    'django',
]
PACKAGES = ['django_aggtrigg']

REQUIREMENTS = [
    'Django<1.7',
    'setuptools',
    'psycopg2',
    'jinja2'
]
ENTRY_POINTS = {}
CMDCLASS = {}
SETUP_REQUIREMENTS = [
    'setuptools'
]


# Tox integration.
from setuptools.command.test import test as TestCommand


class Tox(TestCommand):
    """Test command that runs tox."""
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox  # import here, cause outside the eggs aren't loaded.
        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


CMDCLASS['test'] = Tox


if __name__ == '__main__':  # Do not run setup() when we import this module.
    setup(
        name=NAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=README,
        classifiers=CLASSIFIERS,
        keywords=' '.join(KEYWORDS),
        author=AUTHOR,
        author_email=EMAIL,
        url=URL,
        license=LICENSE,
        packages=PACKAGES,
        include_package_data=True,
        zip_safe=False,
        install_requires=REQUIREMENTS,
        entry_points=ENTRY_POINTS,
        cmdclass=CMDCLASS,
        setup_requires=SETUP_REQUIREMENTS,
        repository="http://127.0.0.1"
    )



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of django-json-dbindex nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#


def parse_kwarg(arg):
    """Parse a string and build the corresponding SQL filter
    """
    ops = {'=': '=',
           'gt': '>',
           'gte': '>=',
           'lt': '<',
           'lte': '<='}

    cols = arg.split('__')

    nbe = len(cols)

    if nbe == 1:
        sqw = "%s=" % (cols[0])
    if nbe == 2:
        sqw = "%s %s" % (cols[0], ops[cols[1]])
    return (nbe, cols[0], sqw)
